// 2/4 TO DO: popup window should only show up when TRAINING MODE is on
//  ----> check Globals for correct variable
//

package spamfilter;

import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.nio.charset.*;


/**
  * Connection information related to a client-to-proxy or proxy-to-server
  * <code>SocketChannel</code>. Contains a <code>ConnectionData</code> object
  * that holds data needed by both sides of the client-to-proxy-to-server
  * connection, plus methods to read from and write to the associated channel
  * through the the <code>ConnectionData</code> object.
  * @author Jennifer Knutson
  * @see Proxy
  * @see ConnectionData
  */
public class ChannelInfo {	
 
    private boolean isClientChannel;

    private ConnectionData connData; // client/server shared object	    
    private static Charset charset; // character set
    private static CharsetDecoder decoder; // byte-to-char translation
    
    /**
      * Creates a new <code>ChannelInfo</code> object with the connection type
      * specified by the <code>isClient</code> flag. Generally, in the case of a
      * client-proxy-server connection, a pair of <code>ChannelInfo</code>
      * objects should be created, with the constructor for each referencing
      * the <b>same</b> <code>ConnectionData</code> object:
      * <p><code>
      * ChannelInfo clientToProxy = new ChannelInfo(shared, true);<br>
      * ChannelInfo proxyToServer = new ChannelInfo(shared, false);<br>
      * </code></p>
      * @param cd a <code>ConnectionData</code> object which is necessary for
      * communication between both halves of a client-proxy-server connection
      * @param isClient the connection type: true for a client-to-proxy
      * connection; false for a proxy-to-server connection.
      * @see ConnectionData
      * @see Proxy
      */
    public ChannelInfo(ConnectionData cd, boolean isClient) {	
	isClientChannel = isClient;
	connData = cd;
	charset = Charset.forName(Globals.charEncoding);
	decoder = charset.newDecoder();
	
	// Set decoder method to IGNORE bad byte sequences in input
	// by dropping the offending character, then continuing
	decoder.onMalformedInput(CodingErrorAction.IGNORE);
    }

    /**
      * Retrieves the appropriate <code>SocketChannel</code> from which to read incoming
      * data, based on the connection type flagged at object creation
      * (client-proxy or a proxy-server).
      * @return an associated <code>SocketChannel</code> from which data can be
      * read
      * @see SocketChannel
      */
    public SocketChannel getReadChannel() {   
	if(isClientChannel)
	    return connData.getClientChannel();	
	else
	    return connData.getServerChannel();	
    }

   /**
     * Retrieves the appropriate <code>SocketChannel</code> to which outgoing data should be
     * written, based on the connection type flagged at object creation
     * (client-proxy or proxy-server).
     * @return an associated <code>SocketChannel</code> to which data can be
     * written
     * @see SocketChannel
     */
    public SocketChannel getWriteChannel() {	
	if(isClientChannel)		
	    return connData.getServerChannel();	
	else		
	    return connData.getClientChannel();		
    }

   
    /**
      * Retrieves the appropriate data-storage <code>StringBuffer</code> for the
      * associated <code>SocketChannel</code>. This buffer resides in the
      * <code>ConnectionData</code> object <code>cd</code>, holding data as it's
      * read from the channel.
      * @return the <code>StringBuffer</code> holding data read from the channel
      * associated with an instance of this object
      * @see ConnectionData
      */ 	
    public StringBuffer getStringBuffer() {	
	if(isClientChannel)		
	    return connData.clientToServerBuffer();	
	else		
	    return connData.serverToClientBuffer();	
    }

    /**
      * Adds <code>String</code> data to the end of the associated data-storage
      * <code>StringBuffer</code>
      * @param str data to append to the end of the data-storage buffer
      * @see #getStringBuffer()
      */
    public void append(String str) {
	getStringBuffer().append(str);			
    }

    /**
      * Reads as much data as possible from the associated
      * <code>SocketChannel</code>'s input stream and store it in the associated
      * <code>StringBuffer</code>; data is then passed on for further analysis.
      * This method detects when a client or server has disconnected, and thus
      * will close both sides of the client-proxy-server connection when
      * necessary.
      * @see #releaseCommands()
      * @see #getReadChannel()
      * @see #getStringBuffer()
      * @see #closeConnection()
      * @throws IOException when the associated channel has been disconnected or
      * there's an error reading data from that channel
      * @throws InterruptedException when the OS sends an interrupt while
      * attempting a read
      */	
    public void read() throws IOException, InterruptedException {
	ByteBuffer buffer = ByteBuffer.allocate(Globals.connectionBufferSize);
	try {
	    int i = getReadChannel().read(buffer);
	    if(i < 0) { // i = number of bytes read
		System.out.println("Channel has been disconnected");
		closeConnection();
		return;
	    }
	    if(i == 0) // nothing read, but connection still good
		return;
	    
	    buffer.flip(); // necessary to read buffer correctly
   
	    // add read, decoded data to channel's buffer
	    append(decoder.decode(buffer).toString());
	    
	    // analyze/pass on any completed lines/commands in buffer
	    // (if client commands not blocked by the Analyzer)
	    if(!(isClientChannel && connData.analyzer().blockClientRequests()))
		releaseCommands();	
	    
	} catch (IOException e) {	   
	    System.err.println("Channel has been disconnected");
	    closeConnection();	    
	    return;
	}
    }

    /**
      * Writes data to the output stream of the associated
      * <code>SocketChannel</code>. 
      * @param str the <code>String</code> data to be written to the output
      * stream
      * @throws IOException on an error when attempting to write to the output
      * stream
      * @see #getWriteChannel()
      */	
    public void write(String str) throws IOException {	
	writeToChannel(getWriteChannel(), str);	
    }


    /**
      * Writes data to the channel connecting the proxy to the mail server.
      * @param str the <code>String</code> data to be written to the mail server
      * @throws IOException on an error when attempting to send data to the mail
      * server
      */
    public void writeToServer(String str) throws IOException {
	if(isClientChannel)
	    write(str);
	else
	    writeToChannel(getReadChannel(), str);
    }

    public void writeToClient(String str) throws IOException {
	if (isClientChannel)
	    writeToChannel(getReadChannel(), str);
	else
	    write(str);

    }

    private void writeToChannel(SocketChannel ch, String str) 
	throws IOException {

	ByteBuffer buffer = ByteBuffer.wrap(str.getBytes(Globals.charEncoding));

	ch.write(buffer);
    }
    

    /**
      * Break up the associated data-storage <code>StringBuffer</code> into
      * distinct lines (commands) to be processed by an <code>Analyzer</code>,
      * which is provided by <code>cd</code>.
      * @see #getStringBuffer()
      * @see ConnectionData
      */
    // break buffer up into distinct lines, then process	
    public void releaseCommands() throws IOException {

	StringBuffer sb = getStringBuffer();
	String cmd = null;
	int i = sb.indexOf(Globals.CRLF);

	while(i >= 0) {		
	    cmd = sb.substring(0, i+2);	

	    // Pass command to Analyzer
	    connData.analyzer().process(cmd, this);	   
	    
	    // Delete the command we just got from the buffer
	    sb = sb.delete(0, i+2);	
	    i = sb.indexOf(Globals.CRLF); // get next command in buffer	
	}
    }
   
    /**
      * Close both sides (<code>SocketChannels</code>) of this
      * client-proxy-server connection by calling <code>ConnectionData</code>'s
      * <code>closeConnection</code> method.
      * @throws IOException on an error closing either channel
      * @see ConnectionData
      */
    public void closeConnection() throws IOException {	
	connData.closeConnection();

	if(Globals.trainMode)
	    connData.analyzer().showFilteredMessages();	
    }

    public boolean isClient() {
	return isClientChannel;
    }

    public boolean isServer() {
	return !isClientChannel;
    }
}
