package spamfilter;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;


// shared info between server-proxy & proxy-client connections
/**
  * Shared information necessary for a client-mailserver connection through a
  * proxy. Keeps track of client-proxy and proxy-server
  * <code>SocketChannels</code>, storing client/server commands in two internal
  * <code>StringBuffers</code>. An <code>Analyzer</code> is provided for
  * processing client/server commands. Methods to read from and write to the
  * channels are not implemented, since the <code>read</code> and
  * <code>write</code> methods are already provided by
  * <code>SocketChannel</code>.
  * @see Proxy
  * @see Analyzer
  * @author Jennifer Knutson
  */
public class ConnectionData {

    private SocketChannel client = null; // channel from client to proxy
    private SocketChannel server = null; // channel from proxy to server	
    private StringBuffer clientToServer = null;
    private StringBuffer serverToClient = null;
    private Analyzer az = null;
    
    /**
      * Creates a connection between a client-to-proxy and a proxy-to-mailserver
      * channel. Sets up buffers to store data being passed through the proxy,
      * as well as an <code>Analyzer</code> for client/server command
      * processing. The exact implementation of <code>Analyzer</code> depends on
      * the protocol being used by the proxy server (for example, POP or IMAP).
      * @param clientChannel the <code>SocketChannel</code> connecting the
      * client to the proxy server 
      * @param serverChannel the <code>SocketChannel</code> connecting the proxy
      * server to the mail server
      * @see Globals#serverType
      * @see SocketChannel
      */
    public ConnectionData(SocketChannel clientChannel, SocketChannel
	    serverChannel) {
	client = clientChannel;
	server = serverChannel;
	clientToServer = new StringBuffer(256);
	serverToClient = new StringBuffer(256);
	if(Globals.serverType.equals("POP"))	
	    az = new POPAnalyzer();
	else	
	    az = new IMAPAnalyzer();
    }		

    /**
      * Retrieves the <code>SocketChannel</code> connecting the client to the
      * proxy server.
      * @return the <code>SocketChannel</code> connecting the client to the
      * proxy server
      * @see SocketChannel
      */		
    public SocketChannel getClientChannel() {
	return client;
    }

    /**
      * Retrieves the <code>SocketChannel</code> connecting the proxy server to
      * the mail server.
      * @return the <code>SocketChannel</code> connecting the proxy server to
      * the mail server
      * @see SocketChannel
      */
    public SocketChannel getServerChannel() {
	return server;	
    }


    /** Retrieves the <code>StringBuffer</code> where data passing from client
      * to server is stored.
      * @return the <code>StringBuffer</code> containing data originating from
      * the client
      */
    public StringBuffer clientToServerBuffer() {
	return clientToServer;
    }


    /**
      * Retrieves the <code>StringBuffer</code> where data passing from server
      * to client is stored.
      * @return the <code>StringBuffer</code> containing data originating from
      * the server
      */
    public StringBuffer serverToClientBuffer() {		
	return serverToClient;	
    }


    /**
      * Retrieves the associated <code>Analyzer</code> which can be used to
      * process client/server commands according to the proxy server's mail
      * protocol. Returns an object of a class that extends
      * <code>Analyzer</code>, based on the proxy's protocol. For example, the
      * retreived <code>Analyzer</code> may be a <code>POPAnalyzer</code> for a
      * proxy serving POP connections, or an <code>IMAPAnalyzer</code> for a
      * proxy serving IMAP connections.
      * @return an <code>Analyzer</code> for processing client/server commands
      * according to the protocol used by the proxy server
      * @see Analyzer
      */
    public Analyzer analyzer() {
	return az;
    }
	
    /**
      * Closes both the client-to-proxy and proxy-to-server
      * <code>SocketChannels</code> associated with an instance of this object.
      * Effectively closes the complete client-to-server connection, so no
      * sockets are left open.
      * @throws IOException when an error occurs closing either the
      * client-to-proxy or the proxy-to-server <code>SocketChannel</code>
      * @see SocketChannel
      */
    public void closeConnection() throws IOException {	
	client.close();
	server.close();
    }
}
