// TO DO:
// * Have block that prints "Proxy Caught an IO Exception" actually
//   notify the GUI that it has shut down unexpectedly, so the start/stop
//   button can be updated appropriately
// * Check to make sure mailserver is a valid hostname

package spamfilter;

import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.*;

/** 
 * An email proxy server which uses non-blocking IO. Currently supports
 * POP or IMAP connections.
 * @author Jennifer Knutson
 */  
public class Proxy extends Thread {
   
    /**
      * Handles demultiplexing of connections
      */
    static Selector selector = null;

    /**
      * Channel which listens for new client connections
      */
    static ServerSocketChannel listeningSock = null;

    // For proper notification of the proxy, so the run() method will return
    private static boolean interrupted = false;

    /**
      * Opens a ServerSocketChannel and an associated Selector to listen for &
      * manage multiple client connections.
      * @param port port on which to listen for client connections
      * @param blockRemoteConnections flag to bind the proxy server to the
      * loopback/local interface only; when false, bind the proxy server to the
      * local machine's IP, opening <code>port</code> for remote connections.
      * @see ServerSocketChannel
      * @see Selector
     */
    public Proxy(int port, boolean blockRemoteConnections) throws IOException {
	try {
	    selector = Selector.open(); // selector manages channels
	    listeningSock = ServerSocketChannel.open();	
	    listeningSock.configureBlocking(false); // non-blocking IO
	
	    if(blockRemoteConnections) // bind only to local/loopback interface
		listeningSock.socket().bind( 
		    new InetSocketAddress("localhost", port));
	    else // determine local IP    
		listeningSock.socket().bind( 
		    new InetSocketAddress(InetAddress.getLocalHost(), port));
	    	
	    System.out.println(Globals.serverType + 
		    " Proxy server listening on port " + port);
	    
	} catch (IOException e) {
	    System.err.print("[Error]: Could not initialize proxy server: ");
	    System.err.println(e.getMessage());

	    try {
		listeningSock.close();
		selector.close();
	    } catch (IOException x) {}
	}
    }
    
  /**
  * Close ServerSocketChannel & associated Selector.
  */  
    private void shutdown() throws IOException {
	System.out.println("Proxy server shutting down.");
	closeAllChannels();
    }	 

    // garbage collector should close sockets in emergency
    public void finalize() throws IOException {
	closeAllChannels();	
    }
    

    // *************************
    // main method of this thread
    public void run() {
	try {
	    interrupted = false;
	    acceptConnections();
	} catch (InterruptedException e) {
	    System.err.println("Proxy was interrupted by the OS: " +
		    e.getMessage());
	} catch (IOException e) {
	    System.err.println("Proxy caught an IO exception: " +
		    e.getMessage());
	}
    }

    // Must call super so the interrupt status of this thread is properly
    // set, thus waking up the selector if it's blocked on selector.select()
    public void interrupt() {
	interrupted = true;
	super.interrupt();
    }

    public int numConnections() {
	// There is always one [accept] key registered with the selector, 
	// since the proxy needs to accept connections; for each additional
	// client connection, 2 more keys are registered. This means that at
	// any given time, ((selector.keys().size - 1) / 2) is the number of 
	// client connections to the proxy
	if(selector == null)
	    return 0;
	else
	    return (selector.keys().size() - 1) / 2;    
    }

    
    private void closeAllChannels() {
	if(selector == null || !selector.isOpen())
	    return;
	    
	SelectionKey key;
	Iterator it = selector.keys().iterator();
	try {
	    while(it.hasNext()) {
		key = (SelectionKey) it.next();
		key.channel().close();
		it.remove(); // got key, so remove it from set
	    }
	    selector.close();  
	} catch (IOException e) {}
    }
    // ****************
    
    
    /**
      * Wait for client connections, accept them, and register them with the
      * selector. Calls the <code>read</code> method of a selection key's
      * attached {@link ChannelInfo} object when the associated SocketChannel
      * becomes readable. Each client-proxy-server connection is managed through
      * two ChannelInfo objects which share a single {@link ConnectionData}
      * object.
      * @see ChannelInfo
      * @see ConnectionData
      */
    public void acceptConnections()     
	throws IOException, InterruptedException {
	    
	    if(listeningSock == null || selector == null || !selector.isOpen())
		return;
	    
	    SelectionKey acceptKey = listeningSock.register(selector, 
		    SelectionKey.OP_ACCEPT);
	    
	    // deal with interesting keys	
	    while(acceptKey.selector().select() > 0 && !interrupted) {
		
		// create iterator to step through set of ready keys
		Iterator it = selector.selectedKeys().iterator();
		
		while(!interrupted && it.hasNext()) {
		    
		    SelectionKey key = (SelectionKey) it.next();		
		    it.remove(); // got key, so remove it from set
					
		    // Client wants to connect: accept connection
		    if(key.isValid() && key.isAcceptable()) {
		    
			ServerSocketChannel proxyChannel =	
			    (ServerSocketChannel) key.channel();
			
			SocketChannel clientChannel = proxyChannel.accept();
			// configure for non-blocking IO		   
			clientChannel.configureBlocking(false);
			
			
			SocketChannel serverChannel = SocketChannel.open(new
			    InetSocketAddress(Globals.mailHost,
				Globals.mailPort));
			// configure for non-blocking IO
			serverChannel.configureBlocking(false);
			
			// register the client & server channels w/ selector
			SelectionKey readFromClient =	
			    clientChannel.register(selector, 
				    SelectionKey.OP_READ);	
			SelectionKey readFromServer =
			    serverChannel.register(selector, 
				    SelectionKey.OP_READ);
	    
			ConnectionData sharedData = new
			    ConnectionData(clientChannel, serverChannel);
			    				    
			// attach channel info/shared object to both the   
			// client & server channels so channels can be	    
			// retrieved when the channel becomes readable
			readFromClient.attach(new 
				ChannelInfo(sharedData, true));
			readFromServer.attach(new 
				ChannelInfo(sharedData, false));	
		    }		
		    else if(key.isValid() && key.isReadable()) {
			((ChannelInfo) key.attachment()).read();
		    }
		}	
	    }
	    
	    // Clean up: the proxy has shut down
	    shutdown();
    }
}    
