/*
* DO WE WANT TO DO TOKEN DEGRADATION? (e.g. subject*FREE -> FREE)?
*
* ALSO:
*    NEED TO HANDLE MISSING PROBABILITIES/COUNT FILES (create new ones)
*    & AUTOMATICALLY BUILD NEW PROBABILITIES TABLE WHEN
*    spamcount & nonspamcount ARE UPDATED!
*
*  IMPORTANT: Get rid of hard-coded filenames!!!!
*
*  NOTE: corpus updating in training mode takes A LONG TIME, not for the actual
*  updating, but for the reading in of the spamcount/nonspamcount files. Could
*  these be loaded at an earlier time, such as when the proxy is started, so
*  there's not such a long delay experienced by the user???
*
*
*  ALSO: the whitelist could be an added pre-layer to any filtering technique
*        but whether or not to use a whitelist should be user-set!!!
*	
*/

package spamfilter;

import java.io.*;
import java.util.*;

// BAYESIAN probability!
/**
 * Checks emails to determine whether they are spam or not, and takes care of 
 * training. Uses a white list and Bayesian filters to determine - based on 
 * previous statistics - what the likelyhood is of an email being spam. This 
 * class also handles training since it already has access to the data needed.
 * @author Tom Ashley and Matt Hansen
 */
public class SpamIdentifier {
    
    private static Email.MailHeaderField mhf = null;
    private static Hashtable tokenList = null;
    private static Hashtable probList = null;
    private static Hashtable whiteList = null;
    private static boolean initialized = false;
    private static boolean WLinitialized = false;
	/**
	 * A sting of token delimitors.
	 */
    private static final String TOKEN_STR = 
	"\t\n\r.!@#%^&*() :;/?<=>[{}]|\\~\"\'";
  
    
    // Load saved hashtable file into memory
    private static void initialize() throws IOException {
	try {    
	    ObjectInputStream in = new ObjectInputStream(new
		    FileInputStream(Globals.probTableFile));
	    probList = (Hashtable) in.readObject();
	    in.close();
 	} catch (ClassNotFoundException e) {
	    System.err.println("Error reading token probabilities file.");
	} catch (FileNotFoundException e) {
            try {
               System.out.println("Prob file missing:  Creating new table");
               ObjectOutputStream out = new ObjectOutputStream(
                  new FileOutputStream(Globals.probTableFile));
               out.writeObject(new Hashtable());
               out.close();
            } catch (IOException ex) {
               System.err.println(ex);
            }
        } catch (IOException e) {
	    System.err.println("*** Warning: error encountered " + 
		    "while initializing probabilities file.");
	}

	if(probList != null) // successfully loaded saved file
	    initialized = true;
    }
	
	 // Load saved White List hashtable file into memory
    private static void WLinitialize() throws IOException {
	try {    
	    ObjectInputStream in2 = new ObjectInputStream(new
		    FileInputStream(Globals.whiteListFile));
	   whiteList = (Hashtable) in2.readObject();
	    in2.close();
 	} catch (ClassNotFoundException e) {
	    System.err.println("Error reading token white list file.");
	} catch (FileNotFoundException e) {
	    try {
               System.out.println("Prob file missing:  Creating new table");
               ObjectOutputStream out = new ObjectOutputStream(
                  new FileOutputStream(Globals.whiteListFile));
               out.writeObject(new Hashtable());
               out.close();
            } catch (IOException ex) {
               System.err.println(ex);
            }
	
	} catch (IOException e) {
	    System.err.println("*** Warning: error encountered " + 
		    "while initializing white list file.");
	}

	if(whiteList != null) // successfully loaded saved file
	    initialized = true;
    }

	/**
	 * Calculates spam probability using Bayesian algorithm. If an address is 
	 * not on the white list, Goes through an email, tokenizes it, and selects 
	 * the 15 most significant tokens based on statistics previously compiled. 
	 * It then applies the Bayesian formula to the probabilities of the 15 most 
	 * significant tokens, and comes up with a probability for the whole email. 
	 * if this probability is below the spam threshhold set by the user, it 
	 * returns <code>false</code> otherwise it returns <code>true</code>.
	 * @param e An Email object.
	 * @return Returns <code>true</code> if the spam probability for Email e 
	 *         is above the spam threshhold (90% by default) or is on the
	 *         white list, else returns <code>false</code>.
	 * @throws IOException if an input or output error occurs.
	 */
    public static boolean isSpam(Email e) throws IOException {
	
	// Load in saved objects from file if it hasn't been done yet
	if(!initialized)
	    initialize();
		
	if(!WLinitialized)
		WLinitialize();

	// This is working now
	if(onWhiteList(e.getSender())) {
               //System.out.println("PASSED BY WHITELIST");
	        return(false);
	}

	// Initialize hashtable for this email's tokens
	tokenList = new Hashtable(2000);
	
	// get email headers, then body & tokenize
	while(e.hasNextHeader()) {
	    mhf = e.getNextHeader();
	    tokenizeProbs(mhf.headerType(), mhf.headerData());
	}
	tokenizeProbs(null, e.getMessageBody());

	// Retrieve the spam probabilities for the tokens in this message
	Vector v = new Vector(tokenList.values());
	// define a way to sort probabilities so that they are sorted
	// by their DISTANCE from 0.5, with the most distant ones first
	Collections.sort(v, new Comparator() {
		public int compare(Object o1, Object o2) {
			float f1 = Math.abs(((Float)o1).floatValue() -
			    (float)0.5);
			float f2 = Math.abs(((Float)o2).floatValue() -
			    (float)0.5);
			return -(new Float(f1)).compareTo(new Float(f2));
		}
			});
	
	
	float result;
	float prod = (float) 1.0;
	float inverseProd = (float) 1.0;
	
	// Apply the Bayesian formula to these probabilities to
	// determine the overall probability that this message is spam
	for(int i = 0; (i < 15 && i < v.size()); i++) {
	    float value = ((Float)v.get(i)).floatValue();
	    
	    //System.out.println((i + 1) + "> " + value);    
	    
	    prod *= value;
	    inverseProd *= (1.0 - value);
	}	
        result = (prod / (prod + inverseProd));

	// debug statement
	System.out.println("SPAM PROBABILITY: " + result);	
	
	if(result >= Globals.spamThreshold) 
	    return true;
	else	    
	    return false;
    }
   
   

    // This function checks a list of safe addresses for users
    private static boolean onWhiteList(String address) {
	
	if(address == null)
	    return false;
	
	int beg = 0;
	int end = 0;
	
	String domainName = null;
	     if((beg = address.indexOf('<')) != -1)
		{
		      end = address.indexOf('>');
	              address = address.substring(beg + 1, end);
		}
            if((beg = address.indexOf('@')) != -1)
            {
                end = address.length();
                domainName = '*' + address.substring(beg,end);
            }

             String temp = (String) whiteList.get(address);
             if(temp != null)
                 return(true);
             else if(domainName != null)
             {
                temp = (String) whiteList.get(domainName);
                if(temp != null)
                   return(true);
             }

	     return(false);	
    }    

    private static String tagToken(String t, String token) {
	if(t != null) {
	    if(t.equalsIgnoreCase("Subject") ||
		    t.equalsIgnoreCase("From") ||
		    t.equalsIgnoreCase("To") ||
		    t.equalsIgnoreCase("Return-Path"))
		token = t.toLowerCase() + "*" + token;	
	    else
		token = "header*" + token; 
	}

	return token;
    }
   
    
    // type = a specific header field (or null for body)
    private static void tokenizeProbs(String headerType, String data) {
	if(data == null)
	    return;
	StringTokenizer st = new StringTokenizer(data, TOKEN_STR);
	
	while(st.hasMoreTokens()) {
	    String token = tagToken(headerType, st.nextToken());
	    Float temp;
	    
	    // put new tokens for this message in tokenList 
	    if(token != null && !tokenList.containsKey(token)) {
		if((temp = (Float) probList.get(token)) != null)
		    tokenList.put(token, temp);		    
		else // give it a neutral (but nonspam-weighted) probability
		    tokenList.put(token, new Float(0.4));
	    }
	}
    }

    private static void tokenizeUpdate(String headerType, String data, 
      Hashtable table) {
         if(data == null)
	     return;
         StringTokenizer st = new StringTokenizer(data, TOKEN_STR);

         while(st.hasMoreTokens()) {
              String token = tagToken(headerType, st.nextToken());
              Integer temp;
            
              if(table.containsKey(token)) {
                   temp = (Integer)table.get(token);
                   table.put(token, new Integer(temp.intValue() + 1));
              }
              else
                   table.put(token, new Integer(1));
         }
    }
   
    /**
	 * Adds the statistics of a vector of previously-identified emails to 
	 * the corpera and updates the probabilities table. Reads in a vector of 
	 * emails and updates the appropriate corpus based on whether the email was 
	 * marked as spam or nonspam. This is generally run after the user ends an 
	 * email session and verifies that the emails were marked correctly. The 
	 * probabilities table is then immediately updated to save the user time on 
	 * startup. 
	 * @param emails A vector of emails - verified as marked correctly by the 
	 *        user - to be trained on.
	 */
    public static void updateCorpus(Vector emails) {
	try {

System.err.println(">>> Now reading in spam/nonspam corpuses");
	    
	    ObjectInputStream inSpam = new ObjectInputStream(new
		    FileInputStream(Globals.spamCorpus));
		int numSpam = ((Integer) inSpam.readObject()).intValue();
	    Hashtable spamCorpus = (Hashtable) inSpam.readObject(); 
		inSpam.close();
		
	    ObjectInputStream inNonSpam = new ObjectInputStream(new
		    FileInputStream(Globals.nonspamCorpus));
	    int numNonSpam = ((Integer) inNonSpam.readObject()).intValue();
	    Hashtable nonSpamCorpus = (Hashtable) inNonSpam.readObject();
	    inNonSpam.close();
		
		Hashtable corpus = null;
            
	    for(int i=0; i < emails.size(); i++) {
	
System.err.println(">>> Using email #" + (i+1) + " to update corpus");
		
		Email e = (Email) emails.get(i);
		if(e.isSpam()) {
		    corpus = spamCorpus;
		    numSpam++;
		}
		else {
		    corpus = nonSpamCorpus;
		    numNonSpam++;
		}
		
		while(e.hasNextHeader()) {
		    Email.MailHeaderField mhf = e.getNextHeader();
		    tokenizeUpdate(mhf.headerType(), mhf.headerData(), corpus);
		}
		tokenizeUpdate(null, e.getMessageBody(), corpus);
	    }
	    
	    ObjectOutputStream outSpam = new ObjectOutputStream(new
		    FileOutputStream(Globals.spamCorpus));
	    outSpam.writeObject(new Integer(numSpam));
	    outSpam.writeObject(spamCorpus);
	    outSpam.close();
	    
	    ObjectOutputStream outNonSpam = new ObjectOutputStream(new
		    FileOutputStream(Globals.nonspamCorpus)); 
	    outNonSpam.writeObject(new Integer(numNonSpam));
	    outNonSpam.writeObject(nonSpamCorpus);
	    outNonSpam.close();
	    
            //Recalculate probabilities based on the new counts
	    GetProb.updateProbCount(spamCorpus, nonSpamCorpus, numSpam, 
				    numNonSpam);

	} catch (ClassNotFoundException e) {
	    System.err.println("Error reading saved corpus file(s).");
	} catch (IOException e) {
	    System.err.println("Gooey Failure, Corpus Update Broken. Boo.");
	}
    } 
}
